/*
HDDVD (EVOB) Demuxer
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define	TRUE		1
#define	FALSE		0

#define	VIDEO		0
#define	AUDIO		1
#define	PADDING	2
#define	OTHER		3
#define	MPEG_AUDIO	4
#define	PRIVATE_1	5
#define	SUBTITLE	6

#define	DD		0
#define	DDPLUS	1
#define	DTSHD		2
#define	LPCM		3
#define	MLP		4

#define	MPEG2		0
#define	H264		1
#define	VC1		2

#define OPTION_CHAR	'-'

unsigned int	frames = 0;

void	parse_mpeg2_video(unsigned char *es_ptr, unsigned int length)
{
	unsigned int	i;
	static unsigned int	parse = 0;

	for(i = 0; i < length; i++)  {
		parse = (parse << 8) + *es_ptr++;
		if (parse == 0x00000100)  {
			frames++;
		}
	}
}

void	parse_h264_video(unsigned char *es_ptr, unsigned int length)
{
	unsigned int	i;
	static unsigned int	parse = 0;

	for(i = 0; i < length; i++)  {
		parse = (parse << 8) + *es_ptr++;
		if (parse == 0x00000109)  {
			frames++;
		}
	}
}

void	parse_vc1_video(unsigned char *es_ptr, unsigned int length)
{
	unsigned int	i;
	static unsigned int	parse = 0;

	for(i = 0; i < length; i++)  {
		parse = (parse << 8) + *es_ptr++;
		if (parse == 0x0000010d)  {
			frames++;
		}
	}
}

void	parse_ddplus_audio(unsigned char *es_ptr, unsigned int length)
{
	unsigned int	i;
	static unsigned int	parse = 0;
	static unsigned int	header_parse = 0;
	static unsigned int	frame_size = 0;
	static int     first_header = FALSE;
	static int     second_header = FALSE;
	static unsigned int	audio_sampling_rate;
	static unsigned int	audio_bitrate;
	static unsigned int	audio_bsid;
	static unsigned int	audio_bsmod;
	static unsigned int	audio_acmod;
	static unsigned int	audio_numblocks;
	static unsigned int	audio_lfeon;
	static unsigned int	eac3_frame_size;

	for(i = 0; i < length; i++)  {
		parse = (parse << 8) + *es_ptr++;
		if ((parse & 0xffff) == 0x00000b77)  {
			if(first_header == FALSE)  {
				header_parse = 5;
				first_header = TRUE;
			}
			else if (second_header == FALSE)  {
				if(frame_size == 5)  {
					second_header = TRUE;
					if (audio_bsid > 10 && audio_bsid <= 16)  {
						printf("Stream is Dolby Digital Plus\n");
						printf("frame_size = %d, number of blocks per frame = %d\n", (eac3_frame_size + 1) * 2, audio_numblocks);
						printf("Audio Bitrate = %d, Audio Sampling Rate = %d\n", audio_bitrate, audio_sampling_rate);
						switch (audio_acmod & 0x7)  {
							case 7:
								printf("Audio Mode = 3/2, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 6:
								printf("Audio Mode = 2/2, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 5:
								printf("Audio Mode = 3/1, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 4:
								printf("Audio Mode = 2/1, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 3:
								printf("Audio Mode = 3/0, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 2:
								printf("Audio Mode = 2/0, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 1:
								printf("Audio Mode = 1/0, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
							case 0:
								printf("Audio Mode = 1+1, bsid = %d, lfeon = %d\n", audio_bsid, audio_lfeon);
								break;
						}
					}
					else  {
						printf("Stream is Dolby Digital\n");
						printf("Audio Bitrate = %d, Audio Sampling Rate = %d\n", audio_bitrate, audio_sampling_rate);
						switch (audio_acmod & 0x7)  {
							case 7:
								printf("Audio Mode = 3/2, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 6:
								printf("Audio Mode = 2/2, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 5:
								printf("Audio Mode = 3/1, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 4:
								printf("Audio Mode = 2/1, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 3:
								printf("Audio Mode = 3/0, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 2:
								printf("Audio Mode = 2/0, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 1:
								printf("Audio Mode = 1/0, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
							case 0:
								printf("Audio Mode = 1+1, bsid = %d, bsmod = %d\n", audio_bsid, audio_bsmod);
								break;
						}
					}
				}
				else  {
					printf("no sync\n");
					first_header = FALSE;
				}
			}
		}
		else if (header_parse != 0)  {
			--header_parse;
			if(header_parse == 4)  {
				eac3_frame_size = (parse & 0x7) << 8;
			}
			else if(header_parse == 3)  {
				eac3_frame_size |= (parse & 0xff);
			}
			else if(header_parse == 2)  {
				switch ((parse & 0xc0) >> 6)  {
					case 3:
						audio_sampling_rate = 0;
						break;
					case 2:
						audio_sampling_rate = 32000;
						break;
					case 1:
						audio_sampling_rate = 44100;
						break;
					case 0:
						audio_sampling_rate = 48000;
						break;
				}
				switch ((parse & 0x3f) >> 1)  {
					case 18:
						audio_bitrate = 640000;
						break;
					case 17:
						audio_bitrate = 576000;
						break;
					case 16:
						audio_bitrate = 512000;
						break;
					case 15:
						audio_bitrate = 448000;
						break;
					case 14:
						audio_bitrate = 384000;
						break;
					case 13:
						audio_bitrate = 320000;
						break;
					case 12:
						audio_bitrate = 256000;
						break;
					case 11:
						audio_bitrate = 224000;
						break;
					case 10:
						audio_bitrate = 192000;
						break;
					case 9:
						audio_bitrate = 160000;
						break;
					case 8:
						audio_bitrate = 128000;
						break;
					case 7:
						audio_bitrate = 112000;
						break;
					case 6:
						audio_bitrate = 96000;
						break;
					case 5:
						audio_bitrate = 80000;
						break;
					case 4:
						audio_bitrate = 64000;
						break;
					case 3:
						audio_bitrate = 56000;
						break;
					case 2:
						audio_bitrate = 48000;
						break;
					case 1:
						audio_bitrate = 40000;
						break;
					case 0:
						audio_bitrate = 32000;
						break;
					default:
						audio_bitrate = 0;
				}
				switch ((parse & 0x30) >> 4)  {
					case 3:
						audio_numblocks = 6;
						break;
					case 2:
						audio_numblocks = 3;
						break;
					case 1:
						audio_numblocks = 2;
						break;
					case 0:
						audio_numblocks = 1;
						break;
				}
				audio_acmod = (parse & 0xe) >> 1;
				audio_lfeon = parse & 0x1;
			}
			else if(header_parse == 1)  {
				audio_bsid = (parse & 0xf8) >> 3;
				audio_bsmod = parse & 0x7;
			}
			else if(header_parse == 0)  {
				if (audio_bsid > 10 && audio_bsid <= 16)  {
					if(audio_sampling_rate == 0)  {
						first_header = FALSE;
					}
					else  {
						frame_size = (eac3_frame_size + 1) * 2;
						audio_bitrate = (frame_size * 8 * (750 / audio_numblocks)) / 4;
					}
				}
				else  {
					audio_acmod = (parse & 0xe0) >> 5;
					if(audio_sampling_rate == 0 || audio_bitrate == 0)  {
						first_header = FALSE;
					}
					else  {
						frame_size = audio_bitrate * 192 / audio_sampling_rate;
					}
				}
			}
		}
		if(first_header == TRUE && second_header == FALSE)  {
			--frame_size;
		}
	}
}

int main(int argc, char **argv)
{
	FILE	*fp, *fpoutv, *fpouta, *fpoutsub;
	static unsigned char	buffer[262144];
	unsigned int	i, j, length;
	static unsigned int	parse = 0;
	static unsigned int	packet_length = 0;
	static unsigned int	packet_length_parse = 0;
	static unsigned int	packet_parse = 0;
	static unsigned int	private_parse = 0;
	static unsigned char	pes_header_flags = 0;
	static unsigned char	pes_header_length = 0;
	static unsigned char	private1_stream_parse = 0;
	static unsigned char	lpcm_stream_parse = 0;
	static unsigned int	mlp_stream_parse = 0;
	static unsigned int	xfer_state = FALSE;
	static unsigned int	stream_type;
	static unsigned int	pts_parse = 0;
	static unsigned int	pts_dts_parse = 0;
	static unsigned int	temp_pts, pts, dts;
	static unsigned int	dump_pts = FALSE;
	static unsigned int	secondary_stream = FALSE;
	static unsigned int	subtitle_stream = FALSE;
	static unsigned int	pes_extension_parse = 0;
	static unsigned int	pes_extension_std_parse = 0;
	static unsigned int	pes_extension_2_parse = 0;
	static unsigned int	pes_extension_flags;
	static unsigned int	pes_extension_2_flags;
	static unsigned int	audio_track;
	static unsigned int	audio_stream_type;
	static unsigned int	audio_stream_id;
	static unsigned int	mpeg2_video_found = FALSE;
	static unsigned int	h264_video_found = FALSE;
	static unsigned int	vc1_video_found = FALSE;
	static unsigned int	dd_audio_found[16];
	static unsigned int	ddplus_audio_found[16];
	static unsigned int	dtshd_audio_found[16];
	static unsigned int	lpcm_audio_found[16];
	static unsigned int	mlp_audio_found[16];
	static unsigned char	mpeg2_video_stream_id = 0xe0;
	static unsigned char	avc_video_stream_id = 0xe2;
	static unsigned char	vc1_video_stream_id = 0x55;
	static unsigned int	subtitle_track;
	static unsigned int	subtitle_found[32];
	static unsigned int	video_stream_type;

	printf("EVOB Demuxer 0.6\n");
	if (argc != 6 && argc != 7 && argc != 9)  {
		fprintf(stderr, "usage: demux <-pst> <infile> <video outfile> <audio outfile> <audio format> <audio track> <subtitle outfile> <subtitle track>\n");
		exit(-1);
	}

	if (argc == 6)  {
		/*--- open binary file (for parsing) ---*/
		fp = fopen(argv[1], "rb");
		if (fp == 0)  {
			fprintf(stderr, "Cannot open bitstream file <%s>\n", argv[1]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpoutv = fopen(argv[2], "wb");
		if (fpoutv == 0)  {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[2]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpouta = fopen(argv[3], "wb");
		if (fpouta == 0)  {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[2]);
			exit(-1);
		}
		if (strcmp(argv[4],"DD") == 0)  {
			audio_stream_type = DD;
		}
		else if (strcmp(argv[4],"dd") == 0)  {
			audio_stream_type = DD;
		}
		else if (strcmp(argv[4],"DD+") == 0)  {
			audio_stream_type = DDPLUS;
		}
		else if (strcmp(argv[4],"dd+") == 0)  {
			audio_stream_type = DDPLUS;
		}
		else if (strcmp(argv[4],"DTS") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[4],"dts") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[4],"DTSHD") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[4],"dtshd") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[4],"LPCM") == 0)  {
			audio_stream_type = LPCM;
		}
		else if (strcmp(argv[4],"lpcm") == 0)  {
			audio_stream_type = LPCM;
		}
		else if (strcmp(argv[4],"MLP") == 0)  {
			audio_stream_type = MLP;
		}
		else if (strcmp(argv[4],"mlp") == 0)  {
			audio_stream_type = MLP;
		}
		else  {
			fprintf(stderr, "<%s> is an invalid audio format\n", argv[4]);
			fprintf(stderr, "try dd, dd+, dtshd, lpcm or mlp\n");
			exit(-1);
		}			
		audio_track = atoi(argv[5]);
	}
	else  {
		if(*argv[1] == OPTION_CHAR)  {
			for(i = 1; i < strlen(argv[1]); i++)  {
				switch (argv[1][i])  {
					case 'p':
					case 'P':
						dump_pts = TRUE;
						break;
					case 's':
					case 'S':
						secondary_stream = TRUE;
						break;
					case 't':
					case 'T':
						subtitle_stream = TRUE;
						break;
					default:
						fprintf(stderr, "Unsupported Option: %c\n", argv[1][i]);
				}
			}
		}
		else  {
			fprintf(stderr, "usage: demux <-pst> <infile> <video outfile> <audio outfile> <audio format> <audio track> <subtitle outfile> <subtitle track>\n");
			exit(-1);
		}
		if (subtitle_stream == TRUE)  {
			if (argc != 9)  {
				fprintf(stderr, "usage: demux <-pst> <infile> <video outfile> <audio outfile> <audio format> <audio track> <subtitle outfile> <subtitle track>\n");
				exit(-1);
			}
		}
		/*--- open binary file (for parsing) ---*/
		fp = fopen(argv[2], "rb");
		if (fp == 0)  {
			fprintf(stderr, "Cannot open bitstream file <%s>\n", argv[2]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpoutv = fopen(argv[3], "wb");
		if (fpoutv == 0)  {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[2]);
			exit(-1);
		}

		/*--- open binary file (for parsing) ---*/
		fpouta = fopen(argv[4], "wb");
		if (fpouta == 0)  {
			fprintf(stderr, "Cannot open output file <%s>\n", argv[2]);
			exit(-1);
		}
		if (strcmp(argv[5],"DD") == 0)  {
			audio_stream_type = DD;
		}
		else if (strcmp(argv[5],"dd") == 0)  {
			audio_stream_type = DD;
		}
		else if (strcmp(argv[5],"DD+") == 0)  {
			audio_stream_type = DDPLUS;
		}
		else if (strcmp(argv[5],"dd+") == 0)  {
			audio_stream_type = DDPLUS;
		}
		else if (strcmp(argv[5],"DTS") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[5],"dts") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[5],"DTSHD") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[5],"dtshd") == 0)  {
			audio_stream_type = DTSHD;
		}
		else if (strcmp(argv[5],"LPCM") == 0)  {
			audio_stream_type = LPCM;
		}
		else if (strcmp(argv[5],"lpcm") == 0)  {
			audio_stream_type = LPCM;
		}
		else if (strcmp(argv[5],"MLP") == 0)  {
			audio_stream_type = MLP;
		}
		else if (strcmp(argv[5],"mlp") == 0)  {
			audio_stream_type = MLP;
		}
		else  {
			fprintf(stderr, "<%s> is an invalid audio format\n", argv[5]);
			fprintf(stderr, "try dd, dd+, dtshd, lpcm or mlp\n");
			exit(-1);
		}			
		audio_track = atoi(argv[6]);
		if (subtitle_stream == TRUE)  {
			/*--- open binary file (for parsing) ---*/
			fpoutsub = fopen(argv[7], "wb");
			if (fpoutsub == 0)  {
				fprintf(stderr, "Cannot open output file <%s>\n", argv[7]);
				exit(-1);
			}
			subtitle_track = atoi(argv[8]);
		}
	}

	if (audio_stream_type == DD)  {
		audio_stream_id = 0x80;
	}
	else if (audio_stream_type == DDPLUS)  {
		audio_stream_id = 0xc0;
	}
	else if (audio_stream_type == DTSHD)  {
		audio_stream_id = 0x88;
	}
	else if (audio_stream_type == LPCM)  {
		audio_stream_id = 0xa0;
	}
	else if (audio_stream_type == MLP)  {
		audio_stream_id = 0xb0;
	}

	if (secondary_stream == TRUE)  {
		mpeg2_video_stream_id = 0xe1;
		avc_video_stream_id = 0xe3;
		vc1_video_stream_id = 0x56;
	}

	for (i = 0; i < 16; i++)  {
		dd_audio_found[i] = FALSE;
		ddplus_audio_found[i] = FALSE;
		dtshd_audio_found[i] = FALSE;
		lpcm_audio_found[i] = FALSE;
	}
	for (i = 0; i < 32; i++)  {
		subtitle_found[i] = FALSE;
	}
	while (!feof(fp))  {
		length = fread(&buffer[0], 1, 262144, fp);
		for (i = 0; i < length; i++)  {
			parse = (parse << 8) + buffer[i];
			if (xfer_state == TRUE)  {
				if (stream_type == VIDEO || stream_type == AUDIO || stream_type == MPEG_AUDIO || stream_type == SUBTITLE)  {
					if ((length - i) >= packet_length)  {
						j = packet_length;
					}
					else  {
						j = length - i;
					}
					if (stream_type == VIDEO)  {
						fwrite(&buffer[i], 1, j, fpoutv);
						if (video_stream_type == MPEG2)  {
							parse_mpeg2_video(&buffer[i], j);
						}
						else if (video_stream_type == H264)  {
							parse_h264_video(&buffer[i], j);
						}
						else if (video_stream_type == VC1)  {
							parse_vc1_video(&buffer[i], j);
						}
					}
					else if (stream_type == AUDIO || stream_type == MPEG_AUDIO) {
						fwrite(&buffer[i], 1, j, fpouta);
						if (audio_stream_type == DDPLUS)  {
							parse_ddplus_audio(&buffer[i], j);
						}
					}
					else if (stream_type == SUBTITLE) {
						fwrite(&buffer[i], 1, j, fpoutsub);
					}
#if 0
					printf("burst length = %d, index = %d\n", j, i);
#endif
					packet_length = packet_length - j;
					i = i + j;
					--i;
				}
				else  {
					if ((length - i) >= packet_length)  {
						j = packet_length;
					}
					else  {
						j = length - i;
					}
					packet_length = packet_length - j;
					i = i + j;
					--i;	/* adjust because of for loop */
				}
				if (packet_length == 0)  {
					xfer_state = FALSE;
				}
			}
			else  {
				if (private_parse != 0)  {
					--packet_length;
					--private_parse;
					switch (private_parse)  {
						case 0:
							break;
					}
				}
				else if (parse == (0x00000100 | mpeg2_video_stream_id))  {
					if (mpeg2_video_found == FALSE)  {
						mpeg2_video_found = TRUE;
						printf("MPEG-2 Video Stream found!\n");
					}
					stream_type = VIDEO;
					video_stream_type = MPEG2;
					packet_length_parse = 2;
				}
				else if (parse == (0x00000100 | avc_video_stream_id))  {
					if (h264_video_found == FALSE)  {
						h264_video_found = TRUE;
						printf("H.264 Video Stream found!\n");
					}
					stream_type = VIDEO;
					video_stream_type = H264;
					packet_length_parse = 2;
				}
				else if (parse == 0x000001fd)  {
					if (vc1_video_found == FALSE)  {
						vc1_video_found = TRUE;
						printf("VC-1 Video Stream found!\n");
					}
					stream_type = VIDEO;
					video_stream_type = VC1;
					packet_length_parse = 2;
				}
				else if ((parse & 0xfffffff0) == (0x000001c0 | audio_track))  {
					stream_type = MPEG_AUDIO;
					packet_length_parse = 2;
				}
				else if (parse == 0x000001bc)  {
					stream_type = OTHER;
					packet_length_parse = 2;
				}
				else if (parse == 0x000001bd)  {
					stream_type = PRIVATE_1;
					packet_length_parse = 2;
				}
				else if (parse == 0x000001be)  {
					stream_type = OTHER;
					packet_length_parse = 2;
				}
				else if (parse == 0x000001bf)  {
					stream_type = OTHER;
					packet_length_parse = 2;
				}
				else if (packet_length_parse == 2)  {
					--packet_length_parse;
				}
				else if (packet_length_parse == 1)  {
					--packet_length_parse;
					packet_length = parse & 0xffff;
#if 0
					printf("packet length = %d, 0x%08x\n", packet_length, parse);
#endif
					if (stream_type != OTHER)  {
						packet_parse = 3;
					}
					else  {
						private_parse = packet_length;
					}
				}
				else if (packet_parse != 0)  {
					--packet_length;
					--packet_parse;
					switch (packet_parse)  {
						case 2:
							break;
						case 1:
							pes_header_flags = parse & 0xff;
							break;
						case 0:
							pes_header_length = parse & 0xff;
							if ((pes_header_flags & 0xc0) == 0x80)  {
								pts_parse = 5;
							}
							if ((pes_header_flags & 0xc0) == 0xc0)  {
								pts_dts_parse = 10;
							}
							if ((pes_header_flags & 0xc1) == 0x1)  {
								pes_extension_parse = 1;
							}
							if (pes_header_length == 0)  {
								if (stream_type == PRIVATE_1)  {
									private1_stream_parse = 4;
								}
								else  {
									xfer_state = TRUE;
								}
							}
							break;
					}
				}
				else if (pts_parse != 0)  {
					--packet_length;
					--pes_header_length;
					--pts_parse;
					switch (pts_parse)  {
						case 4:
							temp_pts = 0;
							temp_pts = temp_pts | ((parse & 0x6) << 29);
							break;
						case 3:
							temp_pts = temp_pts | ((parse & 0xff) << 22);
							break;
						case 2:
							temp_pts = temp_pts | ((parse & 0xfe) << 14);
							break;
						case 1:
							temp_pts = temp_pts | ((parse & 0xff) << 7);
							break;
						case 0:
							temp_pts = temp_pts | ((parse & 0xfe) >> 1);
							pts = temp_pts;
							if (stream_type == VIDEO && dump_pts == TRUE)  {
								printf("video pts = %d\n", pts);
							}
							if (pes_header_length == 0)  {
								if (stream_type == PRIVATE_1)  {
									private1_stream_parse = 4;
								}
								else  {
									xfer_state = TRUE;
								}
							}
							else  {
								if (stream_type == VIDEO)  {
									if (pes_header_flags & 0x1)  {
										pes_extension_parse = 1;
									}
								}
							}
							break;
					}
				}
				else if (pts_dts_parse != 0)  {
					--packet_length;
					--pes_header_length;
					--pts_dts_parse;
					switch (pts_dts_parse)  {
						case 9:
							temp_pts = 0;
							temp_pts = temp_pts | ((parse & 0x6) << 29);
							break;
						case 8:
							temp_pts = temp_pts | ((parse & 0xff) << 22);
							break;
						case 7:
							temp_pts = temp_pts | ((parse & 0xfe) << 14);
							break;
						case 6:
							temp_pts = temp_pts | ((parse & 0xff) << 7);
							break;
						case 5:
							temp_pts = temp_pts | ((parse & 0xfe) >> 1);
							pts = temp_pts;
							if (stream_type == VIDEO && dump_pts == TRUE)  {
								printf("video pts = %d\n", pts);
							}
							break;
						case 4:
							temp_pts = 0;
							temp_pts = temp_pts | ((parse & 0x6) << 29);
							break;
						case 3:
							temp_pts = temp_pts | ((parse & 0xff) << 22);
							break;
						case 2:
							temp_pts = temp_pts | ((parse & 0xfe) << 14);
							break;
						case 1:
							temp_pts = temp_pts | ((parse & 0xff) << 7);
							break;
						case 0:
							temp_pts = temp_pts | ((parse & 0xfe) >> 1);
							dts = temp_pts;
							if (pes_header_length == 0)  {
								if (stream_type == PRIVATE_1)  {
									private1_stream_parse = 4;
								}
								else  {
									xfer_state = TRUE;
								}
							}
							else  {
								if (stream_type == VIDEO)  {
									if (pes_header_flags & 0x1)  {
										pes_extension_parse = 1;
									}
								}
							}
							break;
					}
				}
				else if (pes_extension_parse != 0)  {
					--packet_length;
					--pes_header_length;
					--pes_extension_parse;
					switch (pes_extension_parse)  {
						case 0:
							pes_extension_flags = parse & 0xff;
							if (pes_extension_flags & 0x10)  {
								pes_extension_std_parse = 2;
							}
							else if (pes_extension_flags & 0x1)  {
								pes_extension_2_parse = 2;
							}
							break;
					}
				}
				else if (pes_extension_std_parse != 0)  {
					--packet_length;
					--pes_header_length;
					--pes_extension_std_parse;
					switch (pes_extension_std_parse)  {
						case 1:
							break;
						case 0:
							if (pes_extension_flags & 0x1)  {
								pes_extension_2_parse = 2;
							}
							else if (pes_header_length == 0)  {
								xfer_state = TRUE;
							}
							break;
					}
				}
				else if (pes_extension_2_parse != 0)  {
					--packet_length;
					--pes_header_length;
					--pes_extension_2_parse;
					switch (pes_extension_2_parse)  {
						case 1:
							break;
						case 0:
							pes_extension_2_flags = parse & 0x7f;
							if (pes_extension_2_flags == vc1_video_stream_id)  {
								stream_type = VIDEO;
							}
							else  {
								stream_type = PADDING;
							}
							if (pes_header_length == 0)  {
								xfer_state = TRUE;
							}
							break;
					}
				}
				else if (pes_header_length != 0)  {
					--packet_length;
					--pes_header_length;
					switch (pes_header_length)  {
						case 0:
							if (stream_type == PRIVATE_1)  {
								private1_stream_parse = 4;
							}
							else  {
								xfer_state = TRUE;
							}
							break;
					}
				}
				else if (private1_stream_parse != 0)  {
					--packet_length;
					--private1_stream_parse;
					switch (private1_stream_parse)  {
						case 3:
							if ((parse & 0xe0) == 0x20)  {
								if (subtitle_found[parse & 0x1f] == FALSE)  {
									subtitle_found[parse & 0x1f] = TRUE;
									printf("Subtitle Stream number %d found!\n", parse & 0x1f);
								}
								private1_stream_parse = 0;
								if ((subtitle_stream == TRUE) && ((parse & 0x1f) == subtitle_track))  {
									stream_type = SUBTITLE;
								}
								else  {
									stream_type = PADDING;
								}
								xfer_state = TRUE;
							}
							break;
						case 2:
							break;
						case 1:
							break;
						case 0:
#if 0
							printf("parse = 0x%08x\n", parse);
#endif
							if (((parse >> 24) & 0xf8) == 0x80)  {
								if (dd_audio_found[((parse >> 24) & 0x7)] == FALSE)  {
									dd_audio_found[((parse >> 24) & 0x7)] = TRUE;
									printf("DD Audio Stream number %d found!\n", ((parse >> 24) & 0x7));
								}
							}
							else if (((parse >> 24) & 0xf8) == 0x88)  {
								if (dtshd_audio_found[((parse >> 24) & 0x7)] == FALSE)  {
									dtshd_audio_found[((parse >> 24) & 0x7)] = TRUE;
									printf("DTSHD Audio Stream number %d found!\n", ((parse >> 24) & 0x7));
								}
							}
							else if (((parse >> 24) & 0xf0) == 0xc0)  {
								if (ddplus_audio_found[((parse >> 24) & 0xf)] == FALSE)  {
									ddplus_audio_found[((parse >> 24) & 0xf)] = TRUE;
									printf("DD+ Audio Stream number %d found!\n", ((parse >> 24) & 0xf));
								}
							}
							else if (((parse >> 24) & 0xf8) == 0xa0)  {
								if (lpcm_audio_found[((parse >> 24) & 0x7)] == FALSE)  {
									lpcm_audio_found[((parse >> 24) & 0x7)] = TRUE;
									printf("LPCM Audio Stream number %d found!\n", ((parse >> 24) & 0x7));
								}
							}
							else if (((parse >> 24) & 0xf8) == 0xb0)  {
								if (mlp_audio_found[((parse >> 24) & 0x7)] == FALSE)  {
									mlp_audio_found[((parse >> 24) & 0x7)] = TRUE;
									printf("MLP Audio Stream number %d found!\n", ((parse >> 24) & 0x7));
								}
							}
							if (audio_stream_type == MLP)  {
								if ((parse >> 24) == (audio_track | audio_stream_id))  {
									if (dump_pts)  {
										printf("audio pts = %d\n", pts);
									}
									mlp_stream_parse = 1;
								}
								else  {
									stream_type = PADDING;
									xfer_state = TRUE;
								}
							}
							else if (audio_stream_type == LPCM)  {
								if ((parse >> 24) == (audio_track | audio_stream_id))  {
									if (dump_pts)  {
										printf("audio pts = %d\n", pts);
									}
									lpcm_stream_parse = 3;
								}
								else  {
									stream_type = PADDING;
									xfer_state = TRUE;
								}
							}
							else {
								if ((parse >> 24) == (audio_track | audio_stream_id))  {
									if (dump_pts)  {
										printf("audio pts = %d\n", pts);
									}
									stream_type = AUDIO;
								}
								else  {
									stream_type = PADDING;
								}
								xfer_state = TRUE;
							}
							break;
					}
				}
				else if (lpcm_stream_parse != 0)  {
					--packet_length;
					--lpcm_stream_parse;
					switch (lpcm_stream_parse)  {
						case 0:
#if 0
							printf("lpcm parse = 0x%08x\n", parse);
#endif
							stream_type = AUDIO;
							xfer_state = TRUE;
							break;
					}
				}
				else if (mlp_stream_parse != 0)  {
					--packet_length;
					--mlp_stream_parse;
					switch (mlp_stream_parse)  {
						case 0:
#if 0
							printf("mlp parse = 0x%08x\n", parse);
#endif
							stream_type = AUDIO;
							xfer_state = TRUE;
							break;
					}
				}
			}
		}
	}
	fclose(fp);
	fclose(fpoutv);
	fclose(fpouta);
	if (subtitle_stream == TRUE)  {
		fclose(fpoutsub);
	}
	printf("number of video frames demuxed = %d\n", frames);
	return 0;
}
